@@ -1,11 +1,11 @@
-import json
 import hashlib
+import json
 from sys import version_info
+
 IS_PY3 = version_info[0] >= 3
 
 if IS_PY3:
   from urllib.parse import quote
-  import io
 else:
   from urllib import quote
 
@@ -31,12 +31,21 @@ from fds.model.put_object_result import PutObjectResult
 from fds.model.subresource import SubResource
 from fds.model.init_multipart_upload_result import InitMultipartUploadResult
 from fds.model.upload_part_result import UploadPartResult
-from fds.model.fds_lifecycle import FDSLifecycleConfig
+from fds.model.fds_lifecycle import FDSLifecycleConfig, FDSLifecycleRule
+from fds.model.copy_object_result import CopyObjectResult
+from fds.model.timestamp_anti_stealing_link_config import TimestampAntiStealingLinkConfig
 import os
 import sys
 from .utils import uri_to_bucket_and_object, to_json_object
 import logging
 
+from fds.model.upload_part_result_list import UploadPartResultList
+from io import BytesIO
+from io import IOBase
+from io import StringIO
+from time import sleep
+
+
 class GalaxyFDSClient(object):
   '''
   Client for Galaxy FDS Service.
@@ -74,7 +83,8 @@ class GalaxyFDSClient(object):
     if endpoint is None:
       endpoint = self.load_config("xiaomi_fds_endpoint")
     if endpoint is not None and len(endpoint.strip()) == 0:
-      logging.warn("endpoint is set to empty, please check ${XIAOMI_FDS_ENDPOINT} or ${FDS_ENDPOINT} in environ variables, or \"xiaomi_fds_endpoint\" in ~/.config/xiaomi/config")
+      logging.warn(
+        "endpoint is set to empty, please check ${XIAOMI_FDS_ENDPOINT} or ${FDS_ENDPOINT} in environ variables, or \"xiaomi_fds_endpoint\" in ~/.config/xiaomi/config")
     return endpoint
 
   def load_access_key(self):
@@ -86,7 +96,8 @@ class GalaxyFDSClient(object):
     if access_key is None:
       access_key = self.load_config("xiaomi_access_key_id")
     if access_key is not None and len(access_key.strip()) == 0:
-      logging.warn("access_key is set to empty, please check ${XIAOMI_ACCESS_KEY_ID} or ${XIAOMI_ACCESS_KEY} in environ variables, or \"xiaomi_access_key_id\" in ~/.config/xiaomi/config")
+      logging.warn(
+        "access_key is set to empty, please check ${XIAOMI_ACCESS_KEY_ID} or ${XIAOMI_ACCESS_KEY} in environ variables, or \"xiaomi_access_key_id\" in ~/.config/xiaomi/config")
     return access_key
 
   def load_secret_key(self):
@@ -98,7 +109,8 @@ class GalaxyFDSClient(object):
     if secret_key is None:
       secret_key = self.load_config("xiaomi_secret_access_key")
     if secret_key is not None and len(secret_key.strip()) == 0:
-      logging.warn("secret_key is set to empty, please check ${XIAOMI_SECRET_ACCESS_KEY} or ${XIAOMI_SECRET_KEY} in environ variables, or \"xiaomi_secret_access_key\" in ~/.config/xiaomi/config")
+      logging.warn(
+        "secret_key is set to empty, please check ${XIAOMI_SECRET_ACCESS_KEY} or ${XIAOMI_SECRET_KEY} in environ variables, or \"xiaomi_secret_access_key\" in ~/.config/xiaomi/config")
     return secret_key
 
   def load_config(self, config_key):
@@ -219,7 +231,7 @@ class GalaxyFDSClient(object):
         response.status_code, response.content, headers)
       raise GalaxyFDSClientException(message)
 
-  def list_objects(self, bucket_name, prefix = '', delimiter = None, max_keys=None):
+  def list_objects(self, bucket_name, prefix='', delimiter=None, max_keys=None):
     '''
     List all objects in a specified bucket with prefix. If the number of objects
     in the bucket is larger than a threshold, you would get a FDSObjectListing
@@ -233,7 +245,7 @@ class GalaxyFDSClient(object):
     if delimiter is None:
       delimiter = self._delimiter
     uri = '%s%s?prefix=%s&delimiter=%s' % \
-        (self._config.get_base_uri(), bucket_name, prefix, delimiter)
+          (self._config.get_base_uri(), bucket_name, prefix, delimiter)
     if max_keys is not None:
       uri += "&maxKeys=" + str(max_keys)
     response = self._request.get(uri, auth=self._auth)
@@ -245,11 +257,11 @@ class GalaxyFDSClient(object):
       if self._config.debug:
         headers = ' header=%s' % response.headers
       message = 'List objects under bucket %s with prefix %s failed, ' \
-          'status=%s, reason=%s%s' % \
-          (bucket_name, prefix, response.status_code, response.content, headers)
+                'status=%s, reason=%s%s' % \
+                (bucket_name, prefix, response.status_code, response.content, headers)
       raise GalaxyFDSClientException(message)
 
-  def list_trash_objects(self, prefix = '', delimiter=None, max_keys=None):
+  def list_trash_objects(self, prefix='', delimiter=None, max_keys=None):
     '''
     Compared with list_objects, it returns a list of objects in the trash.
     :param prefix: The prefix of bucket_name/object_name.
@@ -272,7 +284,7 @@ class GalaxyFDSClient(object):
     delimiter = previous.delimiter
     marker = previous.next_marker
     uri = "%s%s?prefix=%s&delimiter=%s&marker=%s" % \
-        (self._config.get_base_uri(), bucket_name, prefix, delimiter, marker)
+          (self._config.get_base_uri(), bucket_name, prefix, delimiter, marker)
     response = self._request.get(uri, auth=self._auth)
     if response.status_code == requests.codes.ok:
       objects_list = FDSObjectListing(to_json_object(response.content))
@@ -282,9 +294,9 @@ class GalaxyFDSClient(object):
       if self._config.debug:
         headers = ' header=%s' % response.headers
       message = 'List next batch of objects under bucket %s with prefix %s ' \
-          'and marker %s failed, status=%s, reason=%s%s' % \
-          (bucket_name, prefix, marker, response.status_code, response.content,
-          headers)
+                'and marker %s failed, status=%s, reason=%s%s' % \
+                (bucket_name, prefix, marker, response.status_code, response.content,
+                 headers)
       raise GalaxyFDSClientException(message)
 
   def put_object_with_uri(self, uri, data, metadata=None):
@@ -299,6 +311,73 @@ class GalaxyFDSClient(object):
     self.put_object(bucket_name, object_name, data, metadata)
 
   def put_object(self, bucket_name, object_name, data, metadata=None):
+    '''
+    Put the object to a specified bucket. If a object with the same name already
+    existed, it will be overwritten.
+    :param bucket_name: The name of the bucket to whom the object is put
+    :param object_name: The name of the object to put
+    :param data:        The data to put, bytes or a file like object
+    :param metadata:    The metadata of the object
+    :return: The result of putting action server returns
+    '''
+    part_size = self._config.get_part_size()
+    threshold_size = self._config.get_threshold_size()
+    inputstream = None
+
+    if isinstance(data, bytes):
+      inputstream = BytesIO(data)
+    elif isinstance(data, str):
+      inputstream = StringIO(data)
+    elif not IS_PY3 and isinstance(data, file):
+      inputstream = data
+    elif (isinstance(data, IOBase)):
+      if data.seekable():
+        inputstream = data
+      else:
+        buf = data.readlines()
+        if len(buf) > 0:
+          if isinstance(buf[0], str):
+            inputstream = StringIO(''.join(buf))
+          else:
+            inputstream = BytesIO(b''.join(buf))
+        else:
+          inputstream = StringIO('')
+    else:
+      raise GalaxyFDSClientException("Cannot identify data type")
+
+    pos = inputstream.tell()
+    inputstream.seek(0, 2)
+    filen = inputstream.tell()
+    inputstream.seek(0, pos)
+    if filen < threshold_size:
+      return self.put_object_directly(bucket_name, object_name, inputstream, metadata)
+    else:
+      content = inputstream.read(part_size)
+      upload_token = self.init_multipart_upload(bucket_name, object_name)
+      upload_list = []
+      part_number = 1
+      while (content):
+
+        upload_result = None
+        for i in range(3):
+          try:
+            upload_result = self.upload_part(bucket_name, object_name, upload_token.upload_id,
+                                             part_number, content)
+            upload_list.append(upload_result)
+            break
+          except:
+            logging.warning("upload part %d failed, retry after %d seconds" % (part_number, 3))
+            sleep(3)
+
+        part_number = part_number + 1
+        content = inputstream.read(part_size)
+      upload_part_result = UploadPartResultList({"uploadPartResultList": upload_list})
+      return self.complete_multipart_upload(bucket_name=bucket_name, object_name=object_name,
+                                            upload_id=upload_token.upload_id,
+                                            metadata=metadata,
+                                            upload_part_result_list=json.dumps(upload_part_result))
+
+  def put_object_directly(self, bucket_name, object_name, data, metadata=None):
     '''
     Put the object to a specified bucket. If a object with the same name already
     existed, it will be overwritten.
@@ -309,35 +388,55 @@ class GalaxyFDSClient(object):
     :return: The result of putting action server returns
     '''
     uri = '%s%s/%s' % (self._config.get_upload_base_uri(), bucket_name,
-      object_name)
+                       object_name)
     if metadata is None:
       metadata = FDSObjectMetadata()
+    inputstream = None
+    if isinstance(data, bytes):
+      inputstream = BytesIO(data)
+    elif isinstance(data, str):
+      inputstream = StringIO(data)
+    elif not IS_PY3 and isinstance(data, file):
+      inputstream = data
+    elif (isinstance(data, IOBase)):
+      if data.seekable():
+        inputstream = data
+      else:
+        buf = data.readlines()
+        if len(buf) > 0:
+          if isinstance(buf[0], str):
+            inputstream = StringIO(''.join(buf))
+          else:
+            inputstream = BytesIO(b''.join(buf))
+        else:
+          inputstream = StringIO('')
+    else:
+      raise GalaxyFDSClientException("Cannot identify data type")
+
     if self._config.enable_md5_calculate:
       digest = hashlib.md5()
-      if IS_PY3:
-        if isinstance(data, str):
-          data = data.encode(encoding="UTF-8")
-          digest.update(data)
-        elif data.seekable and data.seekable():
-          pos = data.tell()
-          digest.update(data.read())
-          data.seek(0, pos)
-        else:
-          raise GalaxyFDSClientException("Cannot digest data")
-      else:
-        digest.update(data)
+      pos = inputstream.tell()
+      content = inputstream.read()
+      if IS_PY3 and isinstance(content, str):
+        content = content.encode(encoding="UTF-8")
+      digest.update(content)
+      inputstream.seek(0, pos)
 
       metadata.add_header(Common.CONTENT_MD5, digest.hexdigest())
-
-    response = self._request.put(uri, data=data, auth=self._auth,
-        headers=metadata.metadata)
+    response = None
+    if IS_PY3 and isinstance(inputstream, StringIO):
+      response = self._request.put(uri, data=inputstream.read(), auth=self._auth,
+                                   headers=metadata.metadata)
+    else:
+      response = self._request.put(uri, data=inputstream, auth=self._auth,
+                                   headers=metadata.metadata)
     if response.status_code == requests.codes.ok:
       return PutObjectResult(to_json_object(response.content))
     headers = ""
     if self._config.debug:
       headers = ' header=%s' % response.headers
     message = 'Put object failed, status=%s, reason=%s%s' % (
-        response.status_code, response.content, headers)
+      response.status_code, response.content, headers)
     raise GalaxyFDSClientException(message)
 
   def post_object(self, bucket_name, data, metadata=None):
@@ -355,17 +454,17 @@ class GalaxyFDSClient(object):
     if self._config.enable_md5_calculate:
       digest = hashlib.md5()
       digest.update(data)
-      metadata.add_header(Common.CONTENT_MD5,digest.hexdigest())
+      metadata.add_header(Common.CONTENT_MD5, digest.hexdigest())
 
     response = self._request.post(uri, data=data, auth=self._auth,
-        headers=metadata.metadata)
+                                  headers=metadata.metadata)
     if response.status_code == requests.codes.ok:
       return PutObjectResult(to_json_object(response.content))
     headers = ""
     if self._config.debug:
       headers = ' header=%s' % response.headers
     message = 'Post object failed, status=%s, reason=%s%s' % (
-        response.status_code, response.content, headers)
+      response.status_code, response.content, headers)
     raise GalaxyFDSClientException(message)
 
   def get_object_with_uri(self, uri, position=0, size=4096):
@@ -379,7 +478,8 @@ class GalaxyFDSClient(object):
     bucket_name, object_name = uri_to_bucket_and_object(uri)
     return self.get_object(bucket_name, object_name, position, size)
 
-  def get_object(self, bucket_name, object_name, position=0, size=4096, stream=None, version_id=None):
+  def get_object(self, bucket_name, object_name, position=0, size=4096, stream=None,
+                 version_id=None):
     '''
     Get a specified object from a bucket.
     :param bucket_name: The name of the bucket from whom to get the object
@@ -392,14 +492,15 @@ class GalaxyFDSClient(object):
     if position < 0:
       raise GalaxyFDSClientException("Seek position should be no less than 0")
     uri = '%s%s/%s' % (self._config.get_download_base_uri(), bucket_name,
-      object_name)
+                       object_name)
     req_params = dict()
     if version_id:
       req_params["versionId"] = version_id
 
     if position > 0:
-      header = {Common.RANGE : 'bytes=%d-' % position}
-      response = self._request.get(uri, auth=self._auth, headers=header, stream=stream, params=req_params)
+      header = {Common.RANGE: 'bytes=%d-' % position}
+      response = self._request.get(uri, auth=self._auth, headers=header, stream=stream,
+                                   params=req_params)
     else:
       response = self._request.get(uri, auth=self._auth, stream=stream, params=req_params)
     if response.status_code == requests.codes.ok or \
@@ -431,29 +532,29 @@ class GalaxyFDSClient(object):
                                  position=offset)
     length_left = length
     if length_left == -1:
-        length_left = IS_PY3 and sys.maxsize or sys.maxint
+      length_left = IS_PY3 and sys.maxsize or sys.maxint
     try:
-        if data_file:
-            with open(data_file, "wb") as f:
-                for chunk in fds_object.stream:
-                    l = min(length_left, len(chunk));
-                    f.write(chunk[0:l])
-                    length_left -= l
-                    if length_left <= 0:
-                        break
-        else:
-            for chunk in fds_object.stream:
-                l = min(length_left, len(chunk))
-                if IS_PY3:
-                  sys.stdout.buffer.write(chunk[0:l])
-                else:
-                  sys.stdout.write(chunk[0:l])
-                length_left -= l
-                if length_left <= 0:
-                    break
-            sys.stdout.flush()
+      if data_file:
+        with open(data_file, "wb") as f:
+          for chunk in fds_object.stream:
+            l = min(length_left, len(chunk));
+            f.write(chunk[0:l])
+            length_left -= l
+            if length_left <= 0:
+              break
+      else:
+        for chunk in fds_object.stream:
+          l = min(length_left, len(chunk))
+          if IS_PY3:
+            sys.stdout.buffer.write(chunk[0:l])
+          else:
+            sys.stdout.write(chunk[0:l])
+          length_left -= l
+          if length_left <= 0:
+            break
+        sys.stdout.flush()
     finally:
-        fds_object.stream.close()
+      fds_object.stream.close()
 
   def does_object_exists(self, bucket_name, object_name):
     '''
@@ -499,8 +600,7 @@ class GalaxyFDSClient(object):
       return response.headers["object-version-id"]
     return None
 
-
-  def _delete_objects_(self, bucket_name, object_names):
+  def delete_objects(self, bucket_name, object_names):
     '''
     Delete specified objects in the bucket
     :param bucket_name:
@@ -527,7 +627,7 @@ class GalaxyFDSClient(object):
     :param object_name: The name of the object
     '''
     uri = '%s%s/%s?restore=' % (self._config.get_base_uri(),
-      bucket_name, object_name)
+                                bucket_name, object_name)
     response = self._request.put(uri, auth=self._auth)
     if response.status_code != requests.codes.ok:
       headers = ""
@@ -545,7 +645,7 @@ class GalaxyFDSClient(object):
     :param dst_object_name: The target name of the object to rename to
     '''
     uri = '%s%s/%s?renameTo=%s' % (self._config.get_base_uri(),
-      bucket_name, src_object_name, dst_object_name)
+                                   bucket_name, src_object_name, dst_object_name)
     response = self._request.put(uri, auth=self._auth)
     if response.status_code != requests.codes.ok:
       headers = ""
@@ -562,10 +662,11 @@ class GalaxyFDSClient(object):
     :param acl:         The grant(ACL) to add
     '''
     uri = '%s%s?%s' % (self._config.get_base_uri(), bucket_name,
-      SubResource.ACL)
+                       SubResource.ACL)
     acp = self._acl_to_acp(acl)
     response = self._request.put(uri, auth=self._auth, data=json.dumps(acp,
-        default=lambda x : x.to_string()))
+                                                                       default=lambda
+                                                                         x: x.to_string()))
     if response.status_code != requests.codes.ok:
       headers = ""
       if self._config.debug:
@@ -581,7 +682,7 @@ class GalaxyFDSClient(object):
     :return: The got access control list
     '''
     uri = '%s%s?%s' % (self._config.get_base_uri(), bucket_name,
-      SubResource.ACL)
+                       SubResource.ACL)
     response = self._request.get(uri, auth=self._auth)
     if response.status_code == requests.codes.ok:
       acp = AccessControlPolicy(to_json_object(response.content))
@@ -606,7 +707,8 @@ class GalaxyFDSClient(object):
       self._config.get_base_uri(), bucket_name, object_name, SubResource.ACL)
     acp = self._acl_to_acp(acl)
     response = self._request.put(uri, auth=self._auth, data=json.dumps(acp,
-        default=lambda x : x.to_string()))
+                                                                       default=lambda
+                                                                         x: x.to_string()))
     if response.status_code != requests.codes.ok:
       headers = ""
       if self._config.debug:
@@ -646,7 +748,7 @@ class GalaxyFDSClient(object):
     '''
     uri = '%s%s/%s?%s' % (
       self._config.get_base_uri(), bucket_name, object_name,
-        SubResource.METADATA)
+      SubResource.METADATA)
     response = self._request.get(uri, auth=self._auth)
     if response.status_code == requests.codes.ok:
       metadata = self._parse_object_metadata_from_headers(response.headers)
@@ -691,7 +793,7 @@ class GalaxyFDSClient(object):
       headers = ""
       if self._config.debug:
         headers = ' header=%s' % response.headers
-      message = 'Refresh object failed, status=%s, reason=%s%s' %(
+      message = 'Refresh object failed, status=%s, reason=%s%s' % (
         response.status_code, response.content, headers)
       raise GalaxyFDSClientException(message)
 
@@ -735,7 +837,7 @@ class GalaxyFDSClient(object):
     '''
     uri = '%s%s/%s?%s%s' % (
       self._config.get_base_uri(), bucket_name, object_name, "uploadId=" +
-        upload_id, "&partNumber=" + str(part_number))
+      upload_id, "&partNumber=" + str(part_number))
     response = self._request.put(uri, auth=self._auth, data=data)
     if response.status_code == requests.codes.ok:
       result = UploadPartResult(to_json_object(response.content))
@@ -749,7 +851,7 @@ class GalaxyFDSClient(object):
       raise GalaxyFDSClientException(message)
 
   def complete_multipart_upload(self, bucket_name, object_name, upload_id,
-    metadata, upload_part_result_list):
+                                metadata, upload_part_result_list):
     '''
     Complete a multipart upload
     :param bucket_name:
@@ -765,7 +867,7 @@ class GalaxyFDSClient(object):
     if metadata is None:
       metadata = FDSObjectMetadata()
     response = self._request.put(uri, auth=self._auth,
-      data=upload_part_result_list, headers=metadata.metadata)
+                                 data=upload_part_result_list, headers=metadata.metadata)
     if response.status_code == requests.codes.ok:
       result = PutObjectResult(to_json_object(response.content))
       return result
@@ -787,7 +889,7 @@ class GalaxyFDSClient(object):
     '''
     uri = '%s%s/%s?%s' % (
       self._config.get_base_uri(), bucket_name, object_name, "uploadId=" +
-        upload_id)
+      upload_id)
     response = self._request.delete(uri, auth=self._auth, data='')
     if response.status_code != requests.codes.ok:
       headers = ""
@@ -798,7 +900,7 @@ class GalaxyFDSClient(object):
       raise GalaxyFDSClientException(message)
 
   def generate_presigned_uri(self, base_uri, bucket_name, object_name,
-                             expiration, http_method = "GET", content_type = None, sub_resources = None):
+                             expiration, http_method="GET", content_type=None, sub_resources=None):
     '''
     Generate a pre-signed uri to share object with the public
     :param base_uri: The base uri of rest server. Use client's default if 'None' pass
@@ -823,9 +925,9 @@ class GalaxyFDSClient(object):
                Common.EXPIRES, str(int(expiration)))
       else:
         uri = '%s%s/%s?%s&%s=%s&%s=%s&' % \
-             (base_uri, bucket_name, object_name, '&'.join(sub_resources), \
-              Common.GALAXY_ACCESS_KEY_ID, self._auth._app_key, \
-              Common.EXPIRES, str(int(expiration)))
+              (base_uri, bucket_name, object_name, '&'.join(sub_resources), \
+               Common.GALAXY_ACCESS_KEY_ID, self._auth._app_key, \
+               Common.EXPIRES, str(int(expiration)))
       headers = None
       if content_type != None and isinstance(content_type, IS_PY3 and str or basestring):
         headers = {Common.CONTENT_TYPE: content_type}
@@ -838,9 +940,9 @@ class GalaxyFDSClient(object):
                 Common.EXPIRES, str(int(expiration)), Common.SIGNATURE, signature)
       else:
         return '%s%s/%s?%s&%s=%s&%s=%s&%s=%s' % \
-              (base_uri, quote(bucket_name), quote(object_name), '&'.join(sub_resources), \
-               Common.GALAXY_ACCESS_KEY_ID, self._auth._app_key, \
-               Common.EXPIRES, str(int(expiration)), Common.SIGNATURE, signature)
+               (base_uri, quote(bucket_name), quote(object_name), '&'.join(sub_resources), \
+                Common.GALAXY_ACCESS_KEY_ID, self._auth._app_key, \
+                Common.EXPIRES, str(int(expiration)), Common.SIGNATURE, signature)
 
     except Exception as e:
       message = 'Wrong expiration given. ' \
@@ -852,7 +954,32 @@ class GalaxyFDSClient(object):
     Generate a URI for downloading object
     '''
     return '%s%s/%s' % (self._config.get_download_base_uri(), bucket_name,
-      object_name)
+                        object_name)
+
+  def copy_object(self, src_bucket_name, src_object_name, dst_bucket_name, dst_object_name):
+    '''
+    Copy src_object_name from src_bucket_name to dst_bucket_name, and rename it to dst_object_name
+    :param src_bucket_name: Source bucket name
+    :param src_object_name: Source object name
+    :param dst_bucket_name: Target bucket name
+    :param dst_object_name: Target object name
+    :return:
+    '''
+    uri = '%s%s/%s?cp=cpparam' % (self._config.get_upload_base_uri(), dst_bucket_name,
+                                  dst_object_name)
+
+    data = {"srcBucketName": src_bucket_name, "srcObjectName": src_object_name}
+
+    response = self._request.put(uri, data=json.dumps(data, default=lambda x: x.to_string()),
+                                 auth=self._auth)
+    if response.status_code == requests.codes.ok:
+      return CopyObjectResult(to_json_object(response.content))
+    headers = ""
+    if self._config.debug:
+      headers = ' header=%s' % response.headers
+    message = 'Copy object failed, status=%s, reason=%s%s' % (
+      response.status_code, response.content, headers)
+    raise GalaxyFDSClientException(message)
 
   def _acp_to_acl(self, acp):
     '''
@@ -869,6 +996,10 @@ class GalaxyFDSClient(object):
       return acl
     return str()
 
+  # def copy_object(self, srcBucketName, srcObjectName, dstBucketName, dstObjectName):
+  #   cp_request = FDSCopyObjectRequest(srcBucketName, srcObjectName, dstBucketName, dstObjectName)
+  #   return self.copy_object(cp_request)
+
   def _acl_to_acp(self, acl):
     '''
     Translate AccessControlList to AccessControlPolicy.
@@ -887,7 +1018,7 @@ class GalaxyFDSClient(object):
     Parse object metadata from the response headers.
     '''
     metadata = FDSObjectMetadata()
-    header_keys = [c.lower() for c in response_headers.keys()];
+    header_keys = [c.lower() for c in response_headers.keys()]
     for key in FDSObjectMetadata.PRE_DEFINED_METADATA:
       if key.lower() in header_keys:
         metadata.add_header(key, response_headers[key])
@@ -896,7 +1027,7 @@ class GalaxyFDSClient(object):
         metadata.add_user_metadata(key, response_headers[key])
     return metadata
 
-  def list_all_objects(self, bucket_name, prefix ='', delimiter = None):
+  def list_all_objects(self, bucket_name, prefix='', delimiter=None):
     '''
     traverse all objects in the bucket
     :param bucket_name:
@@ -973,16 +1104,16 @@ class GalaxyFDSClient(object):
         response.status_code, response.content, headers)
       raise GalaxyFDSClientException(message)
 
-  def _update_lifecycle_config_(self, bucket_name, lifecycle_config):
+  def update_lifecycle_config(self, bucket_name, lifecycle_config):
     '''
     Update lifecycle config of a bucket which determine by the config
     :param lifecycle_config:
     '''
     uri = '%s%s' % (self._config.get_base_uri(), bucket_name)
     response = self._request.put(uri,
-      auth=self._auth,
-      params={"lifecycle": ""},
-      data=json.dumps(lifecycle_config))
+                                 auth=self._auth,
+                                 params={"lifecycle": ""},
+                                 data=json.dumps(lifecycle_config))
     if response.status_code != requests.codes.ok:
       headers = ""
       if self._config.debug:
@@ -991,20 +1122,190 @@ class GalaxyFDSClient(object):
         response.status_code, response.content, headers)
       raise GalaxyFDSClientException(message)
 
-  def _get_lifecycle_config_(self, bucket_name):
+  def update_lifecycle_rule(self, bucket_name, rule):
+    '''
+    Update a TTL rule of specified bucket
+    :param rule: If rule.id is None or not exists in bucket.lifecycle_config, add rule with a generated ID. Otherwise bucket.lifecycle_config[rule.id] will be updated
+    '''
+    uri = '%s%s' % (self._config.get_base_uri(), bucket_name)
+    response = self._request.put(uri,
+                                 auth=self._auth,
+                                 params={"lifecycle": "rule"},
+                                 data=json.dumps(rule))
+    if response.status_code != requests.codes.ok:
+      headers = ""
+      if self._config.debug:
+        headers = ' header=%s' % response.headers
+      message = 'Update bucket lifecycle rule failed, status=%s, reason=%s%s' % (
+        response.status_code, response.content, headers)
+      raise GalaxyFDSClientException(message)
+
+  def get_lifecycle_config(self, bucket_name, rule_id=None):
     '''
     Get lifecycle config of specified bucket
     :param bucket_name:
+    :param rule_id: if rule_id is None, return lifecycle_config of the bucket; otherwise return bucket.lifecycle_config[rule_id], if the rule exists
     :return:
     '''
     uri = '%s%s' % (self._config.get_base_uri(), bucket_name)
-    response = self._request.get(uri, auth=self._auth, params={"lifecycle": ""})
+    response = self._request.get(uri, auth=self._auth,
+                                 params={"lifecycle": rule_id and rule_id or ""})
     if response.status_code == requests.codes.ok:
-      return FDSLifecycleConfig(to_json_object(response.content))
+      js = to_json_object(response.content)
+      if rule_id:
+        return FDSLifecycleRule(js)
+      else:
+        return FDSLifecycleConfig(js)
     else:
       headers = ""
       if self._config.debug:
         headers = ' header=%s' % response.headers
       message = 'Get bucket lifecycle config failed, status=%s, reason=%s%s' % (
         response.status_code, response.content, headers)
-      raise GalaxyFDSClientException(message)
\ No newline at end of file
+      raise GalaxyFDSClientException(message)
+
+  def _enable_auto_convert_webp_(self, bucket_name, is_enable):
+    '''
+    Enable bucket auto convert webp.
+    :param bucket_name:     The name of the bucket
+    '''
+    uri = '%s%s?autoWebp=%s' % (
+    self._config.get_base_uri(), bucket_name, (is_enable and "true" or "false"))
+    response = self._request.put(uri, auth=self._auth)
+    if response.status_code != requests.codes.ok:
+      headers = ""
+      if self._config.debug:
+        headers = ' header=%s' % response.headers
+      message = 'Enable bucket auto convert webp failed, status=%s, reason=%s%s' % (
+        response.status_code, response.content, headers)
+      raise GalaxyFDSClientException(message)
+
+  def _is_enable_auto_convert_webp_(self, bucket_name):
+    '''
+    Check the existence of a specified object.
+    :param bucket_name: The name of the bucket
+    :param object_name: The name of the object to check
+    :return: True if the object exists, otherwise, False
+    '''
+    uri = '%s%s?autoWebp' % (self._config.get_base_uri(), bucket_name)
+    response = self._request.get(uri, auth=self._auth)
+    if response.status_code == requests.codes.ok:
+      content = response.content
+      if IS_PY3:
+        content = content.decode("UTF-8")
+      return content.lower() == "true"
+    elif response.status_code == requests.codes.not_found:
+      return False
+    else:
+      headers = ""
+      if self._config.debug:
+        headers = ' header=%s' % response.headers
+      message = 'Check object existence failed, status=%s, reason=%s%s' % (
+        response.status_code, response.content, headers)
+      raise GalaxyFDSClientException(message)
+
+  def _get_webp_(self, bucket_name, object_name, stream=None):
+    '''
+    Get a specified object from a bucket.
+    :param bucket_name: The name of the bucket from whom to get the object
+    :param object_name: The name of the object to get
+    :param stream:      Set True to enable streaming, otherwise, whole object content is read to memory
+    :return: The FDS object
+    '''
+    uri = '%s%s/%s?f=webp' % (self._config.get_download_base_uri(), bucket_name,
+                              object_name)
+    response = self._request.get(uri, auth=self._auth, stream=stream)
+
+    if response.status_code == requests.codes.ok or \
+        response.status_code == requests.codes.partial:
+      obj = FDSObject()
+      obj.stream = response.iter_content()
+      summary = FDSObjectSummary()
+      summary.bucket_name = bucket_name
+      summary.object_name = object_name
+      summary.size = int(response.headers['content-length'])
+      obj.summary = summary
+      obj.metadata = self._parse_object_metadata_from_headers(response.headers)
+      return obj
+    else:
+      headers = ""
+      if self._config.debug:
+        headers = ' header=%s' % response.headers
+      message = 'Get webp failed, status=%s, reason=%s%s' % (
+        response.status_code, response.content, headers)
+      raise GalaxyFDSClientException(message)
+
+  def generate_anti_stealing_uri(self, bucket_name, object_name, expiration, key):
+    '''
+    Generate anti stealing URI
+    :param bucket_name: The name of the bucket from whom to get the object
+    :param object_name: The name of the object to get
+    :param expiration : The timestamp of milliseconds from 1970-01-01-00:00:00
+    :param key        : The key you set up for the bucket
+    :return: The anti stealing URI
+    '''
+    return self._generate_anti_stealing_uri(self._config.get_base_uri(), bucket_name, object_name,
+                                            expiration, key)
+
+  def generate_anti_stealing_cdn_uri(self, bucket_name, object_name, expiration, key):
+    '''
+    Generate anti stealing CDN URI
+    :param bucket_name: The name of the bucket from whom to get the object
+    :param object_name: The name of the object to get
+    :param expiration : The timestamp of milliseconds from 1970-01-01-00:00:00
+    :param key        : The key you set up for the bucket
+    :return: The anti stealing CDN URI
+    '''
+    return self._generate_anti_stealing_uri(self._config.get_cdn_base_uri(), bucket_name,
+                                            object_name, expiration,
+                                            key)
+
+  def _generate_anti_stealing_uri(self, base_uri, bucket_name, object_name, expiration, key):
+    timestamp = int(expiration) / 1000
+    t = hex(timestamp).lstrip('0x')
+    string_to_cal = key + '/' + bucket_name + '/' + object_name + t
+    m = hashlib.md5()
+    m.update(string_to_cal)
+    sign = m.hexdigest()
+
+    return "%s%s/%s?sign=%s&t=%s" % (base_uri, bucket_name, object_name, sign, t)
+
+  def update_timestamp_anti_stealing_link_config(self, bucket_name, config):
+    uri = '%s%s' % (self._config.get_base_uri(), bucket_name)
+    resp = self._request.put(uri, auth=self._auth, params={'antiStealingLink': ''},
+                             data=json.dumps(config, default=lambda x: x.to_string()))
+
+    if resp.status_code != requests.codes.ok:
+      headers = ""
+      if self._config.debug:
+        headers = ' header=%s' % resp.headers
+      message = 'Update timestamp anti stealing link config failed, status=%s, reason=%s%s' % (
+        resp.status_code, resp.content, headers)
+      raise GalaxyFDSClientException(message)
+
+  def get_timestamp_anti_stealing_link_config(self, bucket_name):
+    uri = '%s%s' % (self._config.get_base_uri(), bucket_name)
+    resp = self._request.get(uri, params={'antiStealingLink': ''}, auth=self._auth)
+
+    if resp.status_code == requests.codes.ok:
+      config = TimestampAntiStealingLinkConfig(to_json_object(resp.content))
+      return config
+
+    headers = ""
+    if self._config.debug:
+      headers = ' header=%s' % resp.headers
+    message = 'Get timestamp anti stealing link config failed, status=%s, reason=%s%s' % (
+      resp.status_code, resp.content, headers)
+    raise GalaxyFDSClientException(message)
+
+  def delete_timestamp_anti_stealing_link_config(self, bucket_name):
+    uri = '%s%s' % (self._config.get_base_uri(), bucket_name)
+    resp = self._request.delete(uri, params={'antiStealingLink': ''}, auth=self._auth)
+
+    if resp.status_code != requests.codes.ok:
+      headers = ""
+      if self._config.debug:
+        headers = ' header=%s' % resp.headers
+      message = 'Delete timestamp anti stealing link config failed, status=%s, reason=%s%s' % (
+        resp.status_code, resp.content, headers)
+      raise GalaxyFDSClientException(message)
